require 'assert/factory'
require 'thread'
require 'mr/factory/model_factory'
require 'mr/factory/read_model_factory'
require 'mr/factory/record_factory'
require 'mr/type_converter'

module MR

  module Factory
    extend Assert::Factory
    extend self

    def new(object_class, *args, &block)
      if object_class < MR::Model
        ModelFactory.new(object_class, *args, &block)
      elsif object_class < MR::ReadModelStruct
        ReadModelFactory.new(object_class, *args, &block)
      elsif object_class < MR::Record
        RecordFactory.new(object_class, *args, &block)
      else
        raise ArgumentError, "takes a MR::Model, MR::Record, or MR::ReadModel"
      end
    end

    def primary_key(identifier = nil)
      identifier    ||= 'MR::Factory'
      @primary_keys ||= {}
      @primary_keys[identifier.to_s] ||= PrimaryKeyProvider.new
      self.type_cast(@primary_keys[identifier.to_s].next, :primary_key)
    end

    def decimal(max = nil)
      self.type_cast(Assert::Factory::Random.float(max), :decimal)
    end

    def timestamp
      self.datetime
    end

    def type_converter
      @type_converter ||= MR::TypeConverter.new
    end

    class PrimaryKeyProvider
      attr_reader :mutex, :current
      def initialize
        @current = 0
        @mutex   = Mutex.new
      end
      def next
        @mutex.synchronize{ @current += 1 }
      end
    end

    # used by factories and stacks to say they can't determine a record class for
    # an association (usually a polymorphic association doesn't have its foreign
    # type attribute set)
    class NoRecordClassError < RuntimeError
      def self.for_association(ar_association)
        owner_record_class = ar_association.owner.class
        association_name   = ar_association.reflection.name
        message = "can't build '#{association_name}' association on " \
                  "#{owner_record_class}"
        if ar_association.reflection.options[:polymorphic]
          foreign_type_attribute = ar_association.reflection.foreign_type
          message += " -- try manually setting it, building it via a stack " \
                     "if you've configured default associations, or setting " \
                     "its '#{foreign_type_attribute}' attribute"
        else
          message += " -- try manually setting it or building it via a stack " \
                     "if you've configured default associations"
        end
        self.new(message)
      end
    end

    NoAssociationError = Class.new(ArgumentError)

  end

end
