require 'much-plugin'

module MR; end
module MR::Factory

  module Config
    include MuchPlugin

    plugin_included do
      include InstanceMethods
    end

    module InstanceMethods

      attr_reader :object_class

      def initialize(object_class)
        @object_class = object_class
      end

      # make the methods lazy-eval'd so factories can be built without paying
      # the performance of them building out their config, this makes it less
      # costly to require real and fake factories together in a test suite
      # (i.e. we don't want to split our real and fake factories into separate
      # files)

      def apply_args(object, args)
        apply_default_args(object)
        apply_args_from_hash(object, args)
        true
      end

      def set_default_args(&block)
        @default_args_proc = block
      end

      private

      def object_builder_class; ObjectBuilder; end

      def apply_default_args(object)
        apply_args_from_proc(object, &@default_args_proc) if @default_args_proc
      end

      def apply_args_from_hash(object, args_hash)
        object_builder_class.apply_hash(object, self, args_hash)
      end

      def apply_args_from_proc(object, &args_proc)
        object_builder_class.apply_proc(object, self, &args_proc)
      end

    end

    class ObjectBuilder

      def self.apply_hash(object, factory_config, hash)
        self.new(object, factory_config).tap do |builder|
          hash.each{ |k, v| builder.set(k, v) }
        end
      end

      def self.apply_proc(object, factory_config, &proc)
        self.new(object, factory_config).tap do |builder|
          builder.instance_eval(&proc)
        end
      end

      def initialize(object, factory_config)
        @__mr_ob_object         = object
        @__mr_ob_factory_config = factory_config
      end

      def set(name, value)
        __mr_ob_set_attribute(@__mr_ob_object, name, value)
      rescue ArgumentError => exception
        exception.set_backtrace(caller)
        raise exception
      end

      private

      def __mr_ob_set_attribute(object, name, value)
        object.send("#{name}=", value)
      end
    end

  end

  module WithAssociationsConfig
    include MuchPlugin

    plugin_included do
      include MR::Factory::Config
      include InstanceMethods
    end

    module InstanceMethods

      def record_class
        self.object_class
      end

      def force_in_stack_association_names
        @force_in_stack_association_names ||= []
      end

      def force_in_stack_association?(association_name)
        self.force_in_stack_association_names.include?(association_name)
      end

      def default_factories
        @default_factories ||= Hash.new do |h, record_class|
          # raise a no record class error if passed a `nil` record class, this
          # can happen when a polymorphic association is passed to `factory_for`
          # and its foreign type attribute isn't set
          raise NoRecordClassError if record_class.nil?
          h[record_class] = build_factory_for_record_class(record_class)
        end
      end

      def association_factories
        @association_factories ||= {}
      end

      def factories_for(association_name, record_class)
        key = AssociationFactoriesKey.new(association_name, record_class)
        self.association_factories[key] || [self.default_factories[record_class]]
      end

      def factory_for(association_name, record_class)
        self.factories_for(association_name, record_class).sample
      end

      def factory_config_for(association_name, record_class)
        self.factory_for(association_name, record_class).config
      end

      def record_classes_for(association_name)
        self.factories_for(association_name, nil).map(&:record_class)
      rescue NoRecordClassError
        []
      end

      def build_associated_record(association_name, record_class)
        raise NotImplementedError
      end

      def add_association_factory(association_name, factory, options = nil)
        reflection = self.record_class.reflect_on_association(association_name)
        if reflection.nil?
          raise NoAssociationError, "there is no #{association_name.inspect} " \
                                    "association for #{self.record_class.inspect}"
        end

        options ||= {}
        if options[:force_in_stack]
          (self.force_in_stack_association_names << association_name).uniq!
        end

        add_default_factory(factory.record_class, factory)

        # for polymorphic associations add a lookup with no associated record
        # class, this will allow a polymorphic association without their foreign
        # type attribute set to use the configured factories for the association
        if reflection.options[:polymorphic]
          key = AssociationFactoriesKey.new(association_name)
          add_association_factory_by_key(key, factory)
        end

        key = AssociationFactoriesKey.new(association_name, factory.record_class)
        add_association_factory_by_key(key, factory)
      end

      def ar_association_for(object, name)
        raise NotImplementedError
      end

      private

      def object_builder_class; WithAssociationsConfig::ObjectBuilder; end

      def build_factory_for_record_class(record_class)
        raise NotImplementedError
      end

      def add_association_factory_by_key(key, factory)
        self.association_factories[key] = [] if !self.association_factories.key?(key)
        self.association_factories[key] << factory
      end

      def add_default_factory(record_class, factory)
        if !self.default_factories.key?(record_class)
          self.default_factories[record_class] = factory
        end
      end

    end

    class ObjectBuilder < MR::Factory::Config::ObjectBuilder

      def set(name, value)
        if value.kind_of?(Hash) && (association = __mr_ob_get_association(name))
          __mr_ob_set_association_from_hash(association, name, value)
        else
          super(name, value)
        end
      rescue ArgumentError, NoRecordClassError => exception
        exception.set_backtrace(caller)
        raise exception
      end

      private

      def __mr_ob_get_association(name)
        @__mr_ob_factory_config.ar_association_for(@__mr_ob_object, name)
      end

      def __mr_ob_set_association_from_hash(association, name, value)
        factory = begin
          @__mr_ob_factory_config.factory_for(name, association.klass)
        rescue NoRecordClassError
          raise NoRecordClassError.for_association(association)
        end
        associated_object = @__mr_ob_object.send(name) || factory.instance
        self.class.apply_hash(associated_object, factory.config, value)
        __mr_ob_set_attribute(@__mr_ob_object, name, associated_object)
      end

    end

    module AssociationFactoriesKey
      def self.new(association_name, associated_record_class = nil)
        [association_name, associated_record_class].compact.join("-")
      end
    end

  end

end
