require 'mr/factory'
require 'mr/factory/config'
require 'mr/factory/model_stack'
require 'mr/factory/record_factory'
require 'mr/fake_record'

module MR; end
module MR::Factory

  class ModelFactory

    attr_reader :model_class, :record_class, :config

    def initialize(model_class, record_class, &block)
      @model_class  = model_class
      @record_class = record_class
      @config       = Config.new(@record_class)
      self.instance_eval(&block) if block

      @record_factory = MR::Factory::RecordFactory.new(record_class)
    end

    def model(args = nil)
      @model_class.new(@record_factory.record).tap do |model|
        self.config.apply_args(model, args || {})
      end
    end
    alias :instance :model

    def saved_model(args = nil)
      model = self.model(args).tap(&:save)
      model.record.reset_save_called if model.record.kind_of?(MR::FakeRecord)
      model
    end
    alias :saved_instance :saved_model

    def stack(args = nil)
      MR::Factory::ModelStack.new(self.model(args), self.config)
    end
    alias :instance_stack :stack

    def stack_model(args = nil)
      self.stack(args).model
    end

    def saved_stack(args = nil)
      self.stack(args).tap(&:create)
    end

    def saved_stack_model(args = nil)
      self.saved_stack(args).model
    end

    def saved_dependencies_stack(args = nil)
      self.stack(args).tap(&:create_deps)
    end
    alias :saved_deps_stack :saved_dependencies_stack

    def saved_dependencies_stack_model(args = nil)
      self.saved_deps_stack(args).model
    end
    alias :saved_deps_stack_model :saved_dependencies_stack_model

    private

    def default_association(association_name, factory, options = nil)
      self.config.add_association_factory(association_name, factory, options)
    rescue NoAssociationError => exception
      raise exception.class, exception.message, caller
    end

    def default_args(&block)
      self.config.set_default_args(&block)
    end

    class Config
      include MR::Factory::WithAssociationsConfig

      def build_associated_model(association_name, record_class)
        self.factory_for(association_name, record_class).model
      end

      def build_associated_record(association_name, record_class)
        self.build_associated_model(association_name, record_class).record
      end

      def ar_association_for(model, name)
        if (reflection = model.record_class.reflect_on_association(name))
          model.record.association(reflection.name)
        end
      end

      private

      def build_factory_for_record_class(record_class)
        MR::Factory::ModelFactory.new(record_class.model_class, record_class)
      end
    end

  end


end
