require 'mr/factory'
require 'mr/factory/config'
require 'mr/read_model'

module MR; end
module MR::Factory

  class ReadModelFactory

    attr_reader :read_model_class, :config

    def initialize(read_model_class, &block)
      unless read_model_class < MR::ReadModelStruct
        raise ArgumentError, "takes a read model or read model struct"
      end
      @read_model_class = read_model_class
      @config           = Config.new(read_model_class)
      self.instance_eval(&block) if block
    end

    def read_model_data(args = nil)
      Data.new.tap do |read_model_data|
        self.config.apply_args(read_model_data, args || {})
      end
    end

    def read_model(args = nil)
      @read_model_class.new(self.read_model_data(args))
    end
    alias :instance :read_model

    private

    def default_args(&block)
      self.config.set_default_args(&block)
    end

    class Config
      include MR::Factory::Config

      def read_model_class
        self.object_class
      end

      private

      def apply_default_args(data)
        field_defaults.each do |field_name, field_type|
          data[field_name] = MR::Factory.send(field_type)
        end
        json_struct_list_defaults.each do |field_name, factory|
          data[field_name] = Factory.integer(3).times.map{ factory.read_model_data }
        end
        json_struct_obj_defaults.each do |field_name, factory|
          data[field_name] = factory.read_model_data
        end
        super(data)
      end

      def field_defaults
        @field_defaults ||= self.read_model_class.fields.inject({}) do |h, field|
          h.merge!(field.name.to_s => field.type)
        end
      end

      def json_struct_list_defaults
        @json_struct_list_defaults ||= build_defaults_for_json_struct_fields(
          self.read_model_class.json_struct_lists
        )
      end

      def json_struct_obj_defaults
        @json_struct_obj_defaults ||= build_defaults_for_json_struct_fields(
          self.read_model_class.json_struct_objs
        )
      end

      def build_defaults_for_json_struct_fields(fields)
        fields.inject({}) do |h, field|
          factory = MR::Factory::ReadModelFactory.new(field.struct_class)
          h.merge!(field.name.to_s => factory)
        end
      end
    end

    class Data
      def initialize; @hash = {}; end

      def [](key);         @hash[key];         end
      def []=(key, value); @hash[key] = value; end

      def method_missing(method, *args, &block)
        method_string = method.to_s
        if method_string =~ /=\z/ && args.size == 1
          @hash[method_string.gsub('=', '')] = args.first
        else
          super
        end
      end
    end

  end

end
