require 'mr/factory'
require 'mr/fake_record'

module MR; end
module MR::Factory

  class RecordStack

    def self.for_record(record, factory_config)
      self.new(RecordData.new(record), factory_config)
    end

    attr_reader :record_data, :factory_config, :record_lookup
    attr_reader :record_stacks

    def initialize(record_data, factory_config, record_lookup = nil)
      @record_data    = record_data
      @factory_config = factory_config
      @record_lookup  = record_lookup || build_lookup(@record_data)

      @record_stacks = @record_data.association_datas.map do |association_data|
        next unless association_data.required?(factory_config)

        associated_record_data = association_data.get_record_data(
          @factory_config,
          @record_lookup
        )
        associated_factory_config = @factory_config.factory_config_for(
          association_data.name,
          associated_record_data.record_class
        )
        @record_lookup[associated_record_data.record_class] ||= associated_record_data
        @record_data.set_association(association_data.name, associated_record_data)
        MR::Factory::RecordStack.new(
          associated_record_data,
          associated_factory_config,
          @record_lookup
        )
      end.compact
    end

    def record
      self.record_data.record
    end

    def create
      self.create_dependencies
      self.record_data.create_record
      true
    end

    def destroy
      self.record_data.destroy_record
      self.destroy_dependencies
      true
    end

    def create_dependencies
      self.record_stacks.each(&:create)
      self.record_data.refresh_record_associations
      true
    end
    alias :create_deps :create_dependencies

    def destroy_dependencies
      self.record_stacks.each(&:destroy)
      true
    end
    alias :destroy_deps :destroy_dependencies

    private

    def build_lookup(record_data)
      load_preset_associations_into_lookup({}, record_data)
    end

    def load_preset_associations_into_lookup(lookup, record_data)
      record_data.association_datas.select(&:preset?).each do |association_data|
        preset_record_data = association_data.preset_record_data
        lookup[preset_record_data.record_class] ||= preset_record_data
        load_preset_associations_into_lookup(lookup, preset_record_data)
      end
      lookup
    end

    class RecordData

      attr_reader :record, :association_datas

      def initialize(record)
        @record            = record
        @association_datas = build_association_datas(@record)
      end

      def record_class
        @record.class
      end

      def set_association(name, record_data)
        @record.send("#{name}=", record_data.record)
      end

      def create_record
        if @record.new_record?
          @record.save!
          @record.reset_save_called if @record.kind_of?(MR::FakeRecord)
        end
      end

      def destroy_record
        @record.destroy unless @record.destroyed?
      end

      # ensures records have their associations foreign type/key attributes set
      # by re-setting the association to its current value
      def refresh_record_associations
        @association_datas.each do |association_data|
          associated_record = @record.send(association_data.name)
          @record.send("#{association_data.name}=", associated_record)
        end
      end

      private

      def build_association_datas(record)
        record.class.reflect_on_all_associations(:belongs_to).map do |reflection|
          AssociationData.new(record.association(reflection.name))
        end.compact.sort
      end

    end

    class AssociationData

      attr_reader :association, :preset_record_data

      def initialize(association)
        @association = association

        preset_record = association.owner.send("#{self.name}")
        @preset_record_data = RecordData.new(preset_record) if preset_record
      end

      def name
        self.association.reflection.name
      end

      def record_class
        self.association.klass
      end

      def preset?
        !self.preset_record_data.nil?
      end

      def required?(factory_config)
        self.preset?                                      ||
        factory_config.force_in_stack_association?(self.name) ||
        columns_required?(self.association)
      end

      def get_record_data(factory_config, lookup)
        if self.preset?
          self.preset_record_data
        elsif (lookup_record_data = get_lookup_record_data(factory_config, lookup))
          lookup_record_data
        else
          RecordData.new(build_record(factory_config))
        end
      end

      def <=>(other)
        if other.kind_of?(self.class)
          self.name.to_s <=> other.name.to_s
        else
          super
        end
      end

      private

      # get the intersection of the record classes and use a random match to
      # get a record data from the lookup; if their are no matching record
      # classes the intersection returns an empty array which will return a
      # `nil` value from the lookup
      def get_lookup_record_data(factory_config, lookup)
        record_classes = lookup_record_classes(factory_config) & lookup.keys
        lookup[record_classes.sample]
      end

      def lookup_record_classes(factory_config)
        if self.record_class
          [self.record_class]
        else
          factory_config.record_classes_for(self.name)
        end
      end

      def build_record(factory_config)
        factory_config.build_associated_record(self.name, self.record_class)
      rescue NoRecordClassError
        raise NoRecordClassError.for_association(self.association)
      end

      def columns_required?(association)
        column_required?(association.owner, association.reflection.foreign_key)  ||
        column_required?(association.owner, association.reflection.foreign_type)
      end

      def column_required?(record, column_name)
        column = record.column_for_attribute(column_name)
        !(column.nil? || column.null)
      end

    end

  end

end
