require 'active_model'
require 'active_record'
require 'active_record/validations'
require 'much-plugin'
require 'mr/factory'
require 'mr/fake_record/attributes'

module MR; end
module MR::FakeRecord

  module Persistence
    include MuchPlugin

    plugin_included do
      include MR::FakeRecord::Attributes
      extend TransactionMethods
      extend ClassMethods
      include InstanceMethods

      attribute :id, :primary_key
    end

    # this is broken into a separate module so `FakeRecord` can extend it and
    # provide easy access to the `transaction` method (for stubbing, etc)
    module TransactionMethods

      # ActiveRecord methods

      def transaction
        begin
          yield if block_given?
        rescue ActiveRecord::Rollback
          # activerecord swallows rollback exceptions, they are only intended as a
          # mechanism to rollback the transaction
        end
      end

    end

    module ClassMethods

      # ActiveRecord methods

      # this is needed to raise ActiveRecord::RecordInvalid
      def human_attribute_name(attribute, options = {})
        options[:default] || attribute.to_s.split('.').last
      end

    end

    module InstanceMethods

      # ActiveRecord methods

      def save!
        raise ActiveRecord::RecordInvalid.new(self) unless self.valid?
        self.id ||= MR::Factory.primary_key(self.class)
        current_time = CurrentTime.new
        self.created_at ||= current_time if self.respond_to?(:created_at=)
        if self.respond_to?(:updated_at=) && !self.updated_at_changed?
          self.updated_at = current_time
        end
        self.saved_attributes = self.attributes.dup
        @save_called = true
      end

      def destroy
        @destroyed = true
      end

      def transaction(&block)
        self.class.transaction(&block)
      end

      def new_record?
        !self.id
      end

      def destroyed?
        !!@destroyed
      end

      def errors
        @errors ||= ActiveModel::Errors.new(self)
      end

      def valid?
        self.errors.empty?
      end

      # Non-ActiveRecord methods

      def save_called
        @save_called = false if @save_called.nil?
        @save_called
      end

      def reset_save_called
        @save_called = false
      end

    end

    module CurrentTime
      def self.new
        if ActiveRecord::Base.default_timezone == :utc
          Time.now.utc
        else
          Time.now
        end
      end
    end

  end

end
