require 'much-plugin'
require 'mr/json_field'
require 'mr/read_model/data'
require 'mr/type_converter'

module MR; end
module MR::ReadModel

  module Fields
    include MuchPlugin

    plugin_included do
      include MR::ReadModel::Data
      extend ClassMethods
      include InstanceMethods
    end

    module ClassMethods

      def fields
        @fields ||= MR::ReadModel::FieldSet.new
      end

      def field(name, type)
        fields.add(name, type, self)
      rescue InvalidFieldTypeError => exception
        raise ArgumentError, exception.message, caller
      end

      def json_struct_list(name, struct_class)
        field     = JsonStructListField.new(self, name.to_s, struct_class)
        ivar_name = "@#{name}"

        define_method(name) do
          begin
            instance_variable_get(ivar_name) ||
            instance_variable_set(ivar_name, field.reader(self.read_model_data))
          rescue StandardError => exception
            exception.set_backtrace(caller)
            raise exception
          end
        end

        (self.json_struct_lists << field).uniq!
      end

      def json_struct_lists
        @json_struct_lists ||= []
      end

      def json_struct_obj(name, struct_class)
        field     = JsonStructObjField.new(self, name.to_s, struct_class)
        ivar_name = "@#{name}"

        define_method(name) do
          begin
            instance_variable_get(ivar_name) ||
            instance_variable_set(ivar_name, field.reader(self.read_model_data))
          rescue StandardError => exception
            exception.set_backtrace(caller)
            raise exception
          end
        end

        (self.json_struct_objs << field).uniq!
      end

      def json_struct_objs
        @json_struct_objs ||= []
      end

    end

    module InstanceMethods

      def fields
        @fields ||= begin
          fields = self.class.fields.read_all(self.read_model_data)
          ( self.class.json_struct_lists +
            self.class.json_struct_objs
          ).inject(fields) do |h, field|
            h.merge!(field.name => self.send(field.name))
          end
        end
      end

    end

  end

  class FieldSet
    include Enumerable

    def initialize
      @fields = []
    end

    def find(name)
      @fields.detect{ |f| f.name == name.to_s }
    end

    def read_all(data)
      inject({}) do |h, field|
        h.merge(field.name => field.read(data))
      end
    end

    def add(name, type, model_class = nil)
      @fields << Field.new(name, type).tap do |field|
        field.define_on(model_class) if model_class
      end
    end

    def each(&block)
      @fields.each(&block)
    end

  end

  class Field
    attr_reader :name, :type, :value
    attr_reader :method_name, :ivar_name

    def initialize(name, type)
      @name = name.to_s
      @type = type.to_sym
      unless MR::TypeConverter.valid?(@type)
        raise InvalidFieldTypeError, "`#{@type}` is not a valid field type"
      end

      @method_name    = @name
      @ivar_name      = "@#{@name}"
      @attribute_name = @name
      @converter      = nil
    end

    def read(data)
      @converter ||= MR::TypeConverter.new(determine_ar_column_class(data))
      @converter.send(@type, data[@attribute_name])
    end

    def define_on(model_class)
      field = self
      model_class.class_eval do

        define_method(field.method_name) do
          instance_variable_get(field.ivar_name) ||
          instance_variable_set(field.ivar_name, field.read(self.read_model_data))
        end

      end
    end

    private

    def determine_ar_column_class(data)
      data.class.columns.first.class if data.class.respond_to?(:columns)
    end

  end

  class JsonStructField < Struct.new(:read_model_class, :name, :struct_class_name)

    def struct_class
      @struct_class ||= if self.struct_class_name.kind_of?(String)
        names     = self.struct_class_name.to_s.split('::').reject{ |name| name.empty? }
        namespace = self.struct_class_name =~ /^::/ ? Object : self.read_model_class
        names.inject(namespace){ |o, name| o.const_get(name) }
      else
        self.struct_class_name
      end
    end

    def reader(read_model_data)
      json_value = read_model_data[self.name]
      return json_value if json_value.nil?

      if json_value.kind_of?(String)
        begin
          return MR::JsonField.decode(json_value)
        rescue MR::JsonField::InvalidJSONError => exception
          message = "can't decode `#{self.name}` JSON: #{exception.message}"
          raise exception.class, message, exception.backtrace
        end
      else
        return json_value
      end
    end

  end

  class JsonStructListField < JsonStructField

    def reader(read_model_data)
      if !(json_datas = super(read_model_data)).nil?
        json_datas.map{ |data| self.struct_class.new(data) }
      else
        nil
      end
    end

  end

  class JsonStructObjField < JsonStructField

    def reader(read_model_data)
      if !(json_data = super(read_model_data)).nil?
        self.struct_class.new(json_data)
      else
        nil
      end
    end

  end

  InvalidFieldTypeError = Class.new(RuntimeError)

end
