require 'mr/read_model/subquery'

module MR::ReadModel

  module QueryExpression
    def self.new(type, *args, &block)
      if !args.empty?
        StaticQueryExpression.new(type, *args)
      elsif block
        DynamicQueryExpression.new(type, &block)
      else
        raise InvalidQueryExpressionError, "must be passed args or a block"
      end
    end
  end

  class StaticQueryExpression
    attr_reader :type, :args

    def initialize(type, *args)
      @type = type
      @args = args
    end

    # `apply_to` has to take params that it ignores, so it has the same
    # interface as `DynamicQueryExpression` (which actually uses the params)
    def apply_to(relation, params = nil)
      relation.send(@type, *@args)
    end
  end

  class DynamicQueryExpression
    attr_reader :type, :block

    def initialize(type, &block)
      @type  = type
      @block = block
    end

    def apply_to(relation, params)
      relation.send(@type, relation.instance_exec(params, &@block))
    end
  end

  class MergeQueryExpression
    attr_accessor :type, :query_expression

    def initialize(type, *args, &block)
      @type = type
      @query_expression = QueryExpression.new(:merge, *args, &block)
    end

    def apply_to(relation, params = nil)
      @query_expression.apply_to(relation, params)
    end
  end

  class SubqueryExpression
    TYPES = {
      :joins => JoinSubquery
    }.freeze

    attr_reader :subquery_type, :subquery_args, :subquery_block
    alias :type :subquery_type

    def initialize(type, *args, &block)
      raise ArgumentError, "a block must be provided" unless block
      @subquery_type  = type
      @subquery_args  = args
      @subquery_block = block
    end

    def subquery
      @subquery ||= TYPES[self.subquery_type].new(
        *self.subquery_args,
        &self.subquery_block
      )
    end

    def apply_to(relation, params = nil)
      relation.send(self.type, self.subquery.build_sql(params))
    end
  end

  class FromExpression
    attr_reader :record_class

    def initialize(record_class)
      raise ArgumentError, "must be passed a MR::Record" unless record_class < MR::Record
      @record_class = record_class
    end

    def default_find_attr
      @default_find_attr ||= [
        self.record_class.table_name,
        self.record_class.primary_key
      ].join('.')
    end

    def ar_relation(params = nil)
      self.record_class.scoped
    end
  end

  class FromSubqueryExpression
    attr_reader :subquery_block

    def initialize(&block)
      raise ArgumentError, "a block must be provided" unless block
      @subquery_block = block
    end

    # lazy-build `FromSubquery` so its block is not evaluated until the read
    # model `query` method is called, cache it so it doesn't have to be rebuilt
    # for every `query` call
    def from_subquery
      @from_subquery ||= FromSubquery.new(&self.subquery_block)
    end

    def record_class
      self.from_subquery.record_class
    end

    def default_find_attr
      raise NoFindAttrError, "a `find_attr` has to be specified " \
                             "when using a from subquery"
    end

    def ar_relation(params = nil)
      self.record_class.scoped.from(self.from_subquery.build_sql(params))
    end
  end

  class NullFromExpression
    def record_class;              raise NoFromExpressionError; end
    def default_find_attr;         raise NoFromExpressionError; end
    def ar_relation(params = nil); raise NoFromExpressionError; end
  end

  class SetExpression
    OPERATOR_SQL = {
      :union         => 'UNION'.freeze,
      :union_all     => 'UNION ALL'.freeze,
      :intersect     => 'INTERSECT'.freeze,
      :intersect_all => 'INTERSECT ALL'.freeze,
      :except        => 'EXCEPT'.freeze,
      :except_all    => 'EXCEPT ALL'.freeze
    }.freeze

    attr_reader :operator_sql, :read_model_block

    def initialize(operator, &block)
      raise ArgumentError, "a block must be provided" unless block
      @operator_sql     = OPERATOR_SQL[operator]
      @read_model_block = block
    end

    def read_model_class
      @read_model_class ||= begin
        c = Class.new do
          # TODO - fix circular require
          require 'mr/read_model'
          require 'mr/read_model/set_querying'
          include MR::ReadModel
          include MR::ReadModel::SetQuerying
        end
        c.class_eval(&self.read_model_block)
        c
      end
    end

    def combine_sql(sql, params = nil)
      "#{sql} #{self.operator_sql} #{self.read_model_class.build_sql(params)}"
    end
  end

  InvalidQueryExpressionError = Class.new(RuntimeError)

  NoFindAttrError = Class.new(RuntimeError)

  class NoFromExpressionError < RuntimeError
    def initialize
      super "a from expression hasn't been set - " \
            "set one using `from` or `from_subquery`"
    end
  end

end
